/*
 * Deal with file operations
 *
 * Jeremy Fitzharinge <jeremy@softway.oz.au>, May 1993
 */

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/locks.h>
#include <linux/userfs_fs.h>
#include <linux/userfs_fs_sb.h>
#include <linux/userfs_mount.h>
#include <linux/assert.h>
#include <linux/mm.h>
#include <asm/segment.h>

#include "userfs.h"

static int userfs_read(struct inode *inode, struct file *fl, char *buf, int size)
{
	unsigned int max_size;
	up_preamble pre;
	upp_repl repl;
	upp_read_s snd;
	upp_read_r rcv;
	int rd = 0;
	int sz;
	int ret;
		
	if (inode == NULL)
	{
		printk("%s: inode is NULL\n", __FUNCTION__);
		return -EINVAL;
	}

	rcv.data.nelem = 0;
	max_size = 4096 - sizeof_upp_read_r(&rcv);
	
	snd.file = inode->u.userfs_i.handle;
	snd.ctok = fl->u.userfs_f.cred_tok;

	while(size > 0)
	{
		userfs_genpkt(inode->i_sb, &pre, up_read);

		snd.off = fl->f_pos;
		snd.size = MIN(max_size, size);
		
		if ((ret = userfs_doop(inode->i_sb, &pre, &repl,
				       upp_read_s, &snd,
				       upp_read_r, &rcv)) != 0)
		{
			printk("%s: userfs_doop failed %d\n",
			       __FUNCTION__, ret);
			if (rd == 0)
				rd = ret;
			break;
		}

		if (repl.errno != 0)
		{
			if (rd == 0)
				rd = -repl.errno;
			break;
		}

		if (rcv.data.nelem == 0)
			break;

		sz = MIN(size, rcv.data.nelem);

		if (rcv.data.nelem > max_size)
		{
			printk("%s: more than %d bytes in data (!!!) (%d)\n",
			       __FUNCTION__, max_size, rcv.data.nelem);
			sz = MIN(sz, max_size);
		}

		memcpy_tofs(buf, rcv.data.elems, sz);
		fl->f_pos += sz;
		size -= sz;
		rd += sz;
		buf += sz;

		FREE(rcv.data.elems);
	}
	return rd;
}

static int userfs_write(struct inode *inode, struct file *fl, char *buf, int size)
{
	up_preamble pre;
	upp_repl repl;
	upp_write_s snd;
	upp_write_r rcv;
	int wr = 0;
	int sz;
	int ret;
	unsigned char *mbuf = NULL;
	int bufsz = 0;
	unsigned int max_size;
	
	if (inode == NULL)
	{
		printk("%s: inode is NULL\n", __FUNCTION__);
		return -EINVAL;
	}

	snd.data.nelem = 0;
	max_size = 4096 - sizeof_upp_write_s(&snd);
	
	snd.file = inode->u.userfs_i.handle;
	snd.ctok = fl->u.userfs_f.cred_tok;

	while(size > 0)
	{
		sz = MIN(max_size, size);

		if ((mbuf = (unsigned char *)kmalloc(sz, GFP_KERNEL)) == NULL)
		{
			printk("%s: kmalloc(%d) failed\n", __FUNCTION__, sz);
			if (wr == 0)
				wr = -ENOMEM;
			break;
		}
		bufsz = sz;

		memcpy_fromfs(mbuf, buf, sz);

		userfs_genpkt(inode->i_sb, &pre, up_write);

		snd.off = fl->f_pos;
		snd.data.elems = mbuf;
		snd.data.nelem = sz;
		
		if ((ret = userfs_doop(inode->i_sb, &pre, &repl, upp_write_s, &snd, upp_write_r, &rcv)) != 0)
		{
			printk("%s: userfs_doop failed %d\n",
			       __FUNCTION__, ret);
			if (wr == 0)
				wr = ret;
			break;
		}

		kfree_s(mbuf, bufsz);
		mbuf = NULL;
		bufsz = 0;

		if (repl.errno != 0)
		{
			if (wr == 0)
				wr = -repl.errno;
			break;
		}

		sz = rcv.wrote;

		if (sz == 0)
		{
			printk("%s: write op returned 0\n", __FUNCTION__);
			break;
		}
		fl->f_pos += sz;
		size -= sz;
		wr += sz;
		buf += sz;
	}
	if (mbuf != NULL)
	{
		kfree_s(mbuf, bufsz);
		mbuf = NULL;
		bufsz = 0;
	}

	if (wr >= 0)
	{
		inode->i_size = fl->f_pos;
		inode->i_dirt = 1;
	}
	return wr;
}

static int userfs_readdir(struct inode *dir, struct file *fl, struct dirent *dirent, int count)
{
	up_preamble pre;
	upp_repl repl;
	upp_readdir_s snd;
	upp_readdir_r rcv;
	long ret;

	if (!dir || !S_ISDIR(dir->i_mode))
		return -ENOTDIR;

	userfs_genpkt(dir->i_sb, &pre, up_readdir);

	snd.off = fl->f_pos;
	snd.dir = dir->u.userfs_i.handle;
	snd.ctok = fl->u.userfs_f.cred_tok;

	rcv.name.nelem = 0;
	
	if ((ret = userfs_doop(dir->i_sb, &pre, &repl, upp_readdir_s, &snd, upp_readdir_r, &rcv)) != 0)
	{
		printk("%s: userfs_doop failed %d\n", __FUNCTION__, ret);
		return ret;
	}

	if (repl.errno != 0)
	{
		ret = -repl.errno;
		goto out;
	}

	ret = MIN(NAME_MAX, rcv.name.nelem);

	if (ret > 0)
		memcpy_tofs(dirent->d_name, rcv.name.elems, ret);

	put_fs_byte('\0', dirent->d_name+ret);
	put_fs_long(rcv.file.handle, &dirent->d_ino);
	put_fs_word(ret, &dirent->d_reclen);
#if 0
	/* No other FS's seem to set this */
	put_fs_long(fl->f_pos, &dirent->d_off);
#endif
	
	fl->f_pos += rcv.off;

	ret = rcv.off;

out:
	if (rcv.name.nelem != 0)
		FREE(rcv.name.elems);

	return ret;
}

static int userfs_open(struct inode *ino, struct file *file)
{
	upp_open_s snd;
	upp_open_r rcv;
	up_preamble pre;
	upp_repl repl;
	int ret = 0;
	
	userfs_genpkt(ino->i_sb, &pre, up_open);

	userfs_setcred(&snd.cred);
	snd.file = ino->u.userfs_i.handle;

	if ((ret = userfs_doop(ino->i_sb, &pre, &repl,
				upp_open_s, &snd, upp_open_r, &rcv)) != 0)
	{
		printk("%s: userfs_doop failed %d\n", __FUNCTION__, ret);
		goto out;
	}

	if ((ret = -repl.errno) == 0)
		file->u.userfs_f.cred_tok = rcv.ctok;

 out:
	return ret;
}

static void userfs_close(struct inode *ino, struct file *file)
{
	upp_close_s snd;
	up_preamble pre;
	upp_repl repl;
	int ret = 0;
	
	userfs_genpkt(ino->i_sb, &pre, up_close);

	snd.file = ino->u.userfs_i.handle;
	snd.ctok = file->u.userfs_f.cred_tok;

	if ((ret = userfs_doop(ino->i_sb, &pre, &repl,
			       upp_close_s, &snd, void, NULL)) != 0)
		printk("%s: userfs_doop failed %d\n", __FUNCTION__, ret);

}

/* disable things that process won't cope with */
struct file_operations *userfs_probe_fops(struct super_block *sb)
{
	struct file_operations *fops;

	fops = (struct file_operations *)kmalloc(sizeof(*fops), GFP_KERNEL);

	*fops = userfs_file_operations;

	if (!userfs_probe(sb, up_read))
		fops->read = NULL;
	if (!userfs_probe(sb, up_write))
		fops->write = NULL;
	if (!userfs_probe(sb, up_readdir))
		fops->readdir = NULL;
	if (!userfs_probe(sb, up_open))
		fops->open = NULL;
	if (!userfs_probe(sb, up_close))
		fops->release = NULL;
	
	return fops;
}

struct file_operations userfs_file_operations =
{
	NULL,		/* userfs_lseek */
	userfs_read,
	userfs_write,
	userfs_readdir,
	NULL,		/* userfs_select */
	NULL,		/* userfs_ioctl */
	NULL,		/* mmap */
	userfs_open,
	userfs_close,
	NULL		/* can't fsync */
};

