/*
 * Super block/filesystem wide operations
 *
 * Jeremy Fitzharinge <jeremy@softway.oz.au>, May 1993
 */

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/locks.h>
#include <linux/userfs_fs.h>
#include <linux/userfs_fs_sb.h>
#include <linux/userfs_mount.h>
#include <linux/assert.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <asm/segment.h>

#include "userfs.h"

extern int close_fp(struct file *);

/* Close channels, make everything fail later */
void userfs_close_chan(struct super_block *sb)
{
	if (sb->u.userfs_sb.s_toproc)
	{
		close_fp(sb->u.userfs_sb.s_toproc);
		sb->u.userfs_sb.s_toproc = NULL;
	}

	if (sb->u.userfs_sb.s_fromproc)
	{
		close_fp(sb->u.userfs_sb.s_fromproc);
		sb->u.userfs_sb.s_fromproc = NULL;
	}
}

static void userfs_free_super(struct super_block *sb)
{
	lock_super(sb);

	userfs_close_chan(sb);

	if (sb->s_op != &userfs_super_operations)
		kfree_s(sb->s_op, sizeof(struct super_operations));

	kfree_s(sb->u.userfs_sb.s_iops->default_file_ops, sizeof(struct file_operations));
	kfree_s(sb->u.userfs_sb.s_iops, sizeof(struct inode_operations));

	sb->s_dev = 0;

	if (sb->u.userfs_sb.s_nwproc != 0)
	{
		assert(sb->u.userfs_sb.s_wlist != NULL);
		userfs_kiss(sb, UP_WALL, UP_FINISH);
	}

	assert(sb->u.userfs_sb.s_nwproc == 0);
	assert(sb->u.userfs_sb.s_wlist == NULL);
	
	unlock_super(sb);
}
	

static void userfs_put_super(struct super_block *sb)
{
	up_preamble snd;
	upp_repl repl;
	int ret;

	userfs_genpkt(sb, &snd, up_umount);

	if ((ret = userfs_doop(sb, &snd, &repl, void, NULL, void, NULL)) != 0)
		printk("%s: userfs_doop returns %d\n", __FUNCTION__, ret);

	if (repl.errno != 0)
		printk("%s: up_umount failed with %d\n", __FUNCTION__, repl.errno);
	userfs_free_super(sb);
}

static void userfs_put_inode(struct inode *ino)
{
	up_preamble pre;
	upp_iput_s pkt;
	upp_repl repl;
	int ret;

	userfs_genpkt(ino->i_sb, &pre, up_iput);
	pkt.handle = ino->u.userfs_i.handle;

	if ((ret = userfs_doop(ino->i_sb, &pre, &repl, upp_iput_s, &pkt, void, NULL)) != 0)
	{
		printk("%s: userfs_doop failed %d\n", __FUNCTION__, ret);
		return;
	}

	if (repl.errno != 0)
		printk("%s: operation failed with %d\n", __FUNCTION__,
		       repl.errno);
}

static void userfs_statfs_stub(struct super_block *sb, struct statfs *buf)
{
	put_fs_long(USERFS_SUPER_MAGIC, &buf->f_type);
	put_fs_long(1024, &buf->f_bsize);
	put_fs_long(0, &buf->f_blocks);
	put_fs_long(0, &buf->f_bfree);
	put_fs_long(0, &buf->f_bavail);
	put_fs_long(0, &buf->f_files);
	put_fs_long(0, &buf->f_ffree);
	put_fs_long(NAME_MAX, &buf->f_namelen);
	/* Don't know what value to put in buf->f_fsid */
}

static void userfs_statfs(struct super_block *sb, struct statfs *buf)
{
	up_preamble pre;
	upp_repl repl;
	upp_statfs_r rcv;
	int ret;
	struct statfs sfs;
	
	userfs_genpkt(sb, &pre, up_statfs);

	if ((ret = userfs_doop(sb, &pre, &repl,
			       void, NULL, upp_statfs_r, &rcv)) != 0)
	{
		printk("%s: userfs_doop failed: %d\n", __FUNCTION__, ret);
		memset(&sfs, 0, sizeof(struct statfs));
	}
	else
	{
		sfs.f_type = USERFS_SUPER_MAGIC;
		sfs.f_bsize = rcv.bsize;
		sfs.f_blocks = rcv.blocks;
		sfs.f_bfree = rcv.bfree;
		sfs.f_bavail = rcv.bavail;
		sfs.f_files = rcv.files;
		sfs.f_ffree = rcv.ffree;
		sfs.f_fsid = rcv.fsid;
		sfs.f_namelen = rcv.namelen;
	}
	memcpy_tofs(buf, &sfs, sizeof(struct statfs));
}

static inline struct file *fdtofp(int fd)
{
	struct file *file;

	if (fd>=NR_OPEN)
		return NULL;

	file = current->filp[fd];
	
	if (file->f_inode == NULL)
		return NULL;
	
	return file;
}

struct super_block *
userfs_read_super(struct super_block *sb, void *raw_data, int flag)
{
	struct file *toproc;
	struct file *fromproc;
	struct userfs_mount *mntinfo = (struct userfs_mount *)raw_data;
	up_preamble snd;
	upp_mount_r rcv;
	upp_repl repl;
	int ret;
	
	toproc = fromproc = NULL;

	sb->u.userfs_sb.s_toproc = toproc;
	sb->u.userfs_sb.s_fromproc = fromproc;

	if (mntinfo->version != USERFS_VERSION)
	{
		printk("%s: mount data version %d, kernel is %d\n",
		       __FUNCTION__, mntinfo->version, USERFS_VERSION);
		goto fail;
	}

	if ((toproc = fdtofp(mntinfo->fromkern)) == NULL || !(toproc->f_mode&2))
	{
		printk("%s: bad fromkern fd %d\n",
		       __FUNCTION__, mntinfo->fromkern);
		goto fail;
	}
	
	sb->u.userfs_sb.s_toproc = toproc;

	if ((fromproc = fdtofp(mntinfo->tokern)) == NULL || !(fromproc->f_mode&1))
	{
		printk("%s: bad tokern fd %d\n",
		       __FUNCTION__, mntinfo->tokern);
		goto fail;
	}

	sb->u.userfs_sb.s_fromproc = fromproc;
	
	toproc->f_count++;
	fromproc->f_count++;
	
	sb->s_magic = USERFS_SUPER_MAGIC;
	sb->s_blocksize = 2048;
	sb->s_op = &userfs_super_operations;

	sb->u.userfs_sb.s_seq = 0;
	sb->u.userfs_sb.s_iops = &userfs_inode_operations;

	sb->u.userfs_sb.s_nwproc = 0;
	sb->u.userfs_sb.s_canread = 0;
	sb->u.userfs_sb.s_wlist = NULL;
	
	userfs_genpkt(sb, &snd, up_mount);

	ret = userfs_doop(sb, &snd, &repl, void, NULL, upp_mount_r, &rcv);

	if (ret != 0)
	{
		printk("%s: userfs_doop failed with %d\n", __FUNCTION__, ret);
		goto fail;
	}

	if (repl.errno != 0)
	{
		printk("%s: mount failed with %d\n", __FUNCTION__, repl.errno);
		goto fail;
	}

	sb->u.userfs_sb.s_root = rcv.root;

	sb->s_op = userfs_probe_sops(sb);
	sb->u.userfs_sb.s_iops = userfs_probe_iops(sb);

	if (!(sb->s_mounted = iget(sb, sb->u.userfs_sb.s_root.handle))) {
		printk("get root inode failed\n");
		goto fail;
	}

	return sb;
fail:
	printk("%s: failed\n", __FUNCTION__);
	if (sb != NULL)
	{
		sb->s_dev = 0;
		userfs_close_chan(sb);
	}
	return NULL;
}

extern struct inode_operations userfs_inode_operations;

static void
userfs_read_inode(struct inode *inode)
{
	up_preamble pre;
	upp_repl repl;
	upp_iread_s snd;
	upp_iread_r rcv;
	int ret;

	inode->i_op = NULL;
	inode->i_mode = 0;
	
	userfs_genpkt(inode->i_sb, &pre, up_iread);

	snd.handle.handle = inode->i_ino;
	
	if ((ret = userfs_doop(inode->i_sb, &pre, &repl, upp_iread_s, &snd, upp_iread_r, &rcv)) != 0)
	{
		printk("%s: failed to do upp_iread: %d\n", __FUNCTION__, ret);
		return;
	}

	if (repl.errno != 0)
	{
		printk("%s: inode read failed: ino=%d, errno = %d\n",
		       __FUNCTION__, inode->i_ino, repl.errno);
		memset(&rcv, 0, sizeof(rcv));
	}
	else
	{
		assert(rcv.handle.handle == inode->i_ino);
		assert(inode->i_nlink != 0);

		if ((rcv.handle.handle == inode->i_ino) && (inode->i_nlink != 0))
			inode->i_op = inode->i_sb->u.userfs_sb.s_iops;
		else
			memset(&rcv, 0, sizeof(rcv));
	}

	inode->u.userfs_i.handle = rcv.handle;
	inode->i_mode = rcv.ino.mode;
	inode->i_nlink = rcv.ino.nlink;
	inode->i_uid = rcv.ino.uid;
	inode->i_gid = rcv.ino.gid;
	inode->i_size = rcv.ino.size;
	inode->i_atime = rcv.ino.atime;
	inode->i_mtime = rcv.ino.mtime;
	inode->i_ctime = rcv.ino.ctime;
	inode->i_blksize = rcv.ino.blksize;
	inode->i_blocks = rcv.ino.blocks;
	inode->i_rdev = rcv.ino.rdev;

}

static void userfs_write_inode(struct inode *inode)
{
	up_preamble pre;
	upp_iwrite_s snd;
	upp_repl rcv;
	int ret;
	
	userfs_genpkt(inode->i_sb, &pre, up_iwrite);

	snd.handle = inode->u.userfs_i.handle;
	snd.ino.mode = inode->i_mode;
	snd.ino.nlink = inode->i_nlink;
	snd.ino.uid = inode->i_uid;
	snd.ino.gid = inode->i_gid;
	snd.ino.size = inode->i_size;
	snd.ino.atime = inode->i_atime;
	snd.ino.mtime = inode->i_mtime;
	snd.ino.ctime = inode->i_ctime;
	snd.ino.blksize = inode->i_blksize;
	snd.ino.blocks = inode->i_blocks;

	/*
	 * We arn't allowed to fail, so if the IO below doesn't work,
	 * tough.
	 */
	inode->i_dirt = 0;

	if ((ret = userfs_doop(inode->i_sb, &pre, &rcv, upp_iwrite_s, &snd, void, NULL)) != 0)
	{
		printk("%s: transaction failed: %d\n", __FUNCTION__, ret);
		return;
	}

	if (rcv.errno != 0)
		printk("%s: inode write failed: %d\n",
		       __FUNCTION__, rcv.errno);
}

struct super_operations *userfs_probe_sops(struct super_block *sb)
{
	struct super_operations *sops;

	sops = (struct super_operations *)kmalloc(sizeof(*sops), GFP_KERNEL);
	*sops = userfs_super_operations;

	if (!userfs_probe(sb, up_iread))
		sops->read_inode = NULL;
	if (!userfs_probe(sb, up_iwrite))
		sops->write_inode = NULL;
	if (!userfs_probe(sb, up_iput))
		sops->put_inode = NULL;
	if (!userfs_probe(sb, up_umount))
		sops->put_super = userfs_free_super;
	if (!userfs_probe(sb, up_statfs))
		sops->statfs = userfs_statfs_stub;
	
	return sops;
}

struct super_operations userfs_super_operations =
{
	userfs_read_inode,
	NULL,			/* notify_change */
	userfs_write_inode,
	userfs_put_inode,
	userfs_put_super,
	NULL,			/* write_super */
	userfs_statfs,
	NULL			/* remount */
};
