#include "com.h"
#include "symtab.h"
#include "format.h"
#include <getopt.h>

extern int yyparse();

char *prefix = "";

static void preamble(Type *type, const char *ty, int co, const char *name)
{
	format("\n%sunsigned char *%s_%s(%s%s *sp, unsigned char *buf)\n{\n",
	       (type->anon || type->demand) ? "__inline__ static " : "",
	       ty,
	       name, co ? "const " : "", name);
}

static void postamble()
{
	format("return buf;\n}\n");
}

void encode_type(Type *type, const char *name)
{
	Var *var;
	
	assert(type != NULL);

	if (type->nocode || type->coded == 1)
		return;

	type->coded = 1;
	
	switch(type->class)
	{
#define PRE() 	preamble(type, "encode", 1, name)
#define POST()	postamble()
		
	case Typedef:
		encode_type(type->i.typedef_i, type->alias);
		if (strcmp(type->alias, type->i.typedef_i->alias) == 0)
			break;
		PRE();
		format("buf = encode_%s(&sp, buf);\n",
		       type->i.typedef_i->alias);
		POST();
		break;

	case Scalar:
		PRE();
		format("buf = encode_%s(sp, buf);\n",
		       type->alias);
		POST();
		break;

	case Array:
		if (type->norep)
		{
			PRE();
			format("/* No representation */\n");
			POST();
			break;
		}
		encode_type(type->i.array_i, type->i.array_i->alias);
		PRE();
		format("{\nunsigned i;\n\n");
		format("buf = encode_Ulong(&sp->nelem, buf);\n");
		format("for(i = 0; i < sp->nelem; i++) {\n");
		format("buf = encode_%s(&sp->elems[i], buf);\n",
		       type->i.array_i->alias);
		format("}\n}\n");
		POST();
		break;

	case CArray:
		encode_type(type->i.carray_i.type, type->i.carray_i.type->alias);
		PRE();
		format("{\nunsigned i;\n\n");
		format("for(i = 0; i < %d; i++) {\n", type->i.carray_i.size);
		format("buf = encode_%s(sp[i], buf);\n",
		       type->i.carray_i.type->alias);
		format("}\n}\n");
		POST();
		break;
		
	case Struct:
		for(var = type->i.struct_i; var != NULL; var = var->next)
			encode_type(var->type, var->type->alias);
		
		PRE();
		for(var = type->i.struct_i; var != NULL; var = var->next)
		{
			if (var->type->norep)
			{
				format("/* Unrepresentable type */\n");
				continue;
			}
			format("buf = encode_%s(&sp->%s, buf);\n",
			       var->type->alias,
			       var->name);
		}
		POST();
		break;

	case Pointer:
		encode_type(type->i.pointer_i, type->i.pointer_i->alias);
		PRE();
		format("buf = encode_%s(*sp, buf);\n",
		       type->i.pointer_i->alias);
		POST();
		break;

	default:
		printf("#error \"undealt-with type %d\"\n", type->class);
		break;
#undef PRE
#undef POST
	}
}

void decode_type(Type *type, const char *name)
{
	Var *var;
	
	assert(type != NULL);

	if (type->nocode || type->coded == -1)
		return;

	type->coded = -1;

	switch(type->class)
	{
#define PRE()	preamble(type, "decode", 0, name)
#define POST()	postamble()
		
	case Typedef:
		decode_type(type->i.typedef_i, name);
		if (strcmp(type->alias, type->i.typedef_i->alias) == 0)
			break;
		PRE();
		format("buf = decode_%s(&sp, buf);\n",
		       type->i.typedef_i->alias);
		POST();
		break;

	case Scalar:
		PRE();
		format("buf = decode_%s(sp, buf);\n",
		       type->alias);
		break;

	case Array:
		if (type->norep)
		{
			PRE();
			format("/* No representation */\n");
			POST();
			break;
		}
		
		decode_type(type->i.array_i, type->i.array_i->alias);
		PRE();
		format("{\nunsigned i;\n");
		format("buf = decode_Ulong(&sp->nelem, buf);\n");
		format("if (sp->nelem == 0)\n");
		format("{\nsp->elems = 0;\nreturn buf;\n}\n");
		format("sp->elems = (%s *)ALLOC(sizeof(%s) * sp->nelem);\n",
		       type->i.array_i->alias, type->i.array_i->alias);
		format("if (sp->elems == 0) {\nreturn buf;\n}\n");
		format("for(i = 0; i < sp->nelem; i++) {\n");
		format("buf = decode_%s(&sp->elems[i], buf);\n",
		       type->i.array_i->alias);
		format("}\n}\n");
		POST();
		break;

	case CArray:
		decode_type(type->i.carray_i.type, type->i.carray_i.type->alias);
		PRE();
		format("{\nunsigned i;\n\n");
		format("for(i = 0; i < %d; i++) {\n", type->i.carray_i.size);
		format("buf = decode_%s(sp[i], buf);\n",
		       type->i.carray_i.type->alias);
		format("}\n}\n");
		POST();
		break;
		
	case Struct:
		for(var = type->i.struct_i; var != NULL; var = var->next)
			decode_type(var->type, var->type->alias);
		PRE();
		for(var = type->i.struct_i; var != NULL; var = var->next)
		{
			if (var->type->norep)
			{
				format("/* Unrepresentable type */\n");
				continue;
			}
			format("buf = decode_%s(&sp->%s, buf);\n",
			       var->type->alias,
			       var->name);
		}
		POST();
		break;

	case Pointer:
		decode_type(type->i.pointer_i, type->i.pointer_i->alias);
		PRE();
		format("*sp = (%s *)ALLOC(sizeof(%s));\n",
		       type->i.pointer_i->alias, type->i.pointer_i->alias);
		format("if (*sp == NULL) {\nreturn buf;\n}\n");
		format("buf = decode_%s(*sp, buf);\n",
		       type->i.pointer_i->alias);
		POST();
		break;

#undef PRE
#undef POST
	default:
		printf("#error \"undealt-with type %d\"\n", type->class);
		break;
	}
}

static void s_preamble(const Type *type, const char *name)
{
	format("%sunsigned int sizeof_%s(const %s *sp)\n{\n",
	       (type->nocode || type->anon || type->demand) ? "static __inline__ " : "",
	       name, name);
	format("unsigned int sz=0;\n\n");
}

static void s_postamble(void)
{
	format("return sz;\n}\n");
}

void sizeof_type(Type *type, const char *name)
{
	Var *var;
	
	assert(type != NULL);

	if (type->nocode || type->coded == 2)
		return;

	type->coded = 2;
	
	switch(type->class)
	{
#define PRE() 	s_preamble(type, name)
#define POST()	s_postamble()
		
	case Typedef:
		sizeof_type(type->i.typedef_i, type->alias);
		if (strcmp(type->alias, type->i.typedef_i->alias) == 0)
			break;
		PRE();
		format("sz += sizeof_%s(&sp);\n",
		       type->i.typedef_i->alias);
		POST();
		break;

	case Scalar:
		PRE();
		format("sz += sizeof(%s);\n",
		       type->alias);
		POST();
		break;

	case Array:
		if (type->norep)
		{
			PRE();
			format("/* No representation */\n");
			POST();
			break;
		}
		sizeof_type(type->i.array_i, type->i.array_i->alias);
		PRE();
		format("sz += sizeof_Ulong(&sp->nelem);\n");
		format("sz += sizeof_%s(&sp->elems[0]) * sp->nelem;\n",
		       type->i.array_i->alias);
		POST();
		break;

	case CArray:
		sizeof_type(type->i.carray_i.type, type->i.carray_i.type->alias);
		PRE();
		format("sz += sizeof_%s(sp[0])*%d;\n",
		       type->i.carray_i.type->alias, type->i.carray_i.size);
		POST();
		break;
		
	case Struct:
		for(var = type->i.struct_i; var != NULL; var = var->next)
			sizeof_type(var->type, var->type->alias);
		
		PRE();
		for(var = type->i.struct_i; var != NULL; var = var->next)
		{
			if (var->type->norep)
			{
				format("/* Unrepresentable type */\n");
				continue;
			}
			format("sz += sizeof_%s(&sp->%s);\n",
			       var->type->alias,
			       var->name);
		}
		POST();
		break;

	case Pointer:
		sizeof_type(type->i.pointer_i, type->i.pointer_i->alias);
		PRE();
		format("sz += sizeof_%s(*sp);\n",
		       type->i.pointer_i->alias);
		POST();
		break;

	case Void:
		PRE();
		POST();
		break;
		
	default:
		printf("#error \"undealt-with type %d\"\n", type->class);
		break;
#undef PRE
	}
}

typedef void (*Genfunc)(Type *, const char *);

void gen_code(Var *table, Genfunc func)
{
	for(; table != NULL; table = table->next)
	{
		Type *type, *tt;

		type = table->type;

		assert(table->type->class == Typedef);

		tt = type->i.typedef_i;
		
		if (type->demand)
			continue;

		(*func)(type, table->name);
	}
}

extern FILE *yyin;
char *file;

extern int yydebug;

int
main(int argc, char *argv[])
{
	char *cp;
	int och, err = 0;
	int do_encode = 0, do_decode = 0, do_sizeof = 0;
	char cmd[1024];
	
	while((och = getopt(argc, argv, "sedD")) != EOF)
	{
		switch(och)
		{
		case 'd':
			do_decode = 1;
			break;

		case 'e':
			do_encode = 1;
			break;

		case 's':
			do_sizeof = 1;
			break;
			
		case 'D':
			yydebug = 1;
			break;
			
		default:
			err++;
			break;
		}
	}
	
	if (err || optind >= argc)
	{
		fprintf(stderr, "Usage: %s [-des] typedef-file\n", argv[0]);
		exit(1);
	}

	if (!(do_encode || do_decode || do_sizeof))
		fprintf(stderr, "Warning: no functions will be generated\n");
	
	file = strdup(argv[optind]);

	sprintf(cmd, CPP " %s", file);
	if ((yyin = popen(cmd, "r")) == NULL)
	{
		perror("file open failed");
		exit(1);
	}

	if ((cp = strrchr(argv[optind], '/')) == NULL)
		cp = argv[optind];
	else
		cp++;

	prefix = strdup(cp);
	cp = strrchr(prefix, '.');
	if (cp)
		*cp = 0;
	
	init_symtab();

	format("/*\n * DO NOT EDIT BY HAND!\n *\n");
	format(" * This was automatically generated from \"%s\"",
	       argv[optind]);
	format(" by gencode.\n */\n");
	format("#include \"%s.h\"\n", prefix);

	if (yyparse())
	{
		fprintf(stderr, "Bad input file\n");
		exit(1);

	}
	
	global = name_types(global);

	if (do_encode)
	{
		format("\n/*** Type encoding functions ***/\n");
		gen_code(global, encode_type);
	}

	if (do_decode)
	{
		format("\n/*** Type decoding functions ***/\n");
		gen_code(global, decode_type);
	}

	if (do_sizeof)
	{
		format("\n/*** Type Sizeof functions ***/\n");
		gen_code(global, sizeof_type);
	}
	
	exit(0);
}
