#include "com.h"
#include "symtab.h"
#include "format.h"
#include <getopt.h>
#include <ctype.h>

extern int yyparse();

char *prefix;

void dumptype(Type *type, const char *name)
{
	char *str;
	Var *var;
	int ret = 1;
	
	assert(type != NULL);

	if (type->class != Typedef && type->alias != NULL)
	{
		format(type->alias);
		return;
	}

	switch(type->class)
	{
	case Typedef:
		format("typedef ");
		str = type->i.typedef_i->alias;
		type->i.typedef_i->alias = NULL;
		dumptype(type->i.typedef_i, name);
		type->i.typedef_i->alias = str;
		ret = 0;
		break;

	case Scalar:
		if (type->i.scalar_i.Unsigned)
			format("unsigned ");
		switch(type->i.scalar_i.size)
		{
		case Char:	str = "char"; break;
		case Short:	str = "short"; break;
		case Long:	str = "long"; break;
		case Float:	str = "float"; break;
		case Double:	str = "double"; break;
		default:	str = "???"; break;
		}

		format(str);
		break;

	case Array:
		format("struct\n{\n");
		format("Ulong nelem;\n");
		dumptype(mkpointer(type->i.array_i), "elems");
		format(";\n}");
		break;

	case CArray:
		dumptype(type->i.carray_i.type, NULL);
		format(" %s[%d]", name == NULL ? "" : name, type->i.carray_i.size);
		ret = 0;
		break;
		
	case Struct:
		format("struct\n{\n");
		for(var = type->i.struct_i; var != NULL; var = var->next)
		{
			if (var->type->norep)
			{
				format("/* No representation */\n");
				continue;
			}
			dumptype(var->type, NULL);
			format(" %s;\n", var->name);
		}
		format("}");
		break;

	case Pointer:
		dumptype(type->i.pointer_i, NULL);
		format("*");
		break;
		
	default:
		printf("undealt-with type %d", type->class);
		break;
	}
	if (name != NULL && ret)
		format(" %s", name);
}
			
void gen_hdr(Var *table)
{
	format("/* Encoded types */\n\n");
	for(; table != NULL; table = table->next)
	{
		Type *type = table->type;
		
		assert(type->class = Typedef);
		
		if (type->nohdr)
			continue;
		
		dumptype(type, table->name);
		format(";\n");
	}
}

void gen_proto(Var *table)
{
	format("\n/* Type encode/decode routines */\n\n");
	format("#include \"coder.h\"\t/* Standard routines */\n");

	for(; table != NULL; table = table->next)
	{
		Type *type;
		const char *name = table->name;
		const char *st = table->type->nocode ? "static " : "";
		
		assert(table->type->class == Typedef);

		type = table->type->i.typedef_i;
		
		if (type->anon || table->type->demand)
			continue;

		format("%sunsigned char *encode_%s(const %s *, unsigned char *);\n",
		       st, name, name);
		format("%sunsigned char *decode_%s(%s *, unsigned char *);\n",
		       st, name, name);
		format("%sunsigned int sizeof_%s(const %s *);\n\n",
		       st, name, name);
		       
	}
}

extern FILE *yyin;
char *file;

extern int yydebug;

int
main(int argc, char *argv[])
{
	char *upname, *cp;
	char cmd[1024];
	int c, err = 0;

	while((c = getopt(argc, argv, "D")) != EOF)
		switch(c)
		{
		case 'D':
			yydebug = 1;
			break;

		default:
			err++;
			break;
		}
		
	if (err || optind >= argc)
	{
		fprintf(stderr, "Usage: %s typedef-file\n", argv[0]);
		exit(1);
	}

	file = strdup(argv[optind]);
	
	sprintf(cmd, CPP " %s", file);
	if ((yyin = popen(cmd, "r")) == NULL)
	{
		perror("file open failed");
		exit(1);
	}

	if ((cp = strrchr(file, '/')) == NULL)
		cp = file;
	else
		cp++;

	prefix = strdup(cp);
	cp = strrchr(prefix, '.');
	if (cp)
		*cp = 0;

	init_symtab();

	upname = cp = strdup(prefix);

	for(; *cp; cp++)
		if (islower(*cp))
			*cp = toupper(*cp);
	
	format("/*\n * DO NOT EDIT BY HAND!\n *\n");
	format(" * This was automatically generated from \"%s\"",
	       argv[optind]);
	format(" by genhdr.\n */\n");
	format("#ifndef __%s_H_SEEN__\n", upname);
	format("#define __%s_H_SEEN__\n\n", upname);

	if (yyparse())
	{
		fprintf(stderr, "Bad input file\n");
		exit(1);
	}
	
 	global = name_types(global);
	
	gen_hdr(global);
	gen_proto(global);
	
	format("#endif /* __%s_H_SEEN__ */\n", upname);

	exit(0);
}
