/*
 *	Filesystem stub for filesystems implemented with user-level
 *	programs.
 *
 *	This file forms the definition of the protocol between the
 *	kernel and the user process through the file descriptors.
 *
 *	There are two file descriptors passed to the kernel as part
 *	of the args to mount().  One is for data from the kernel
 *	to the filesystem, the other from the filesystem to the
 *	kernel.  Control and data information are encoded along
 *	the same path.
 *
 *	The filesystem process never initiates any transaction - they
 *	are all generated by the kernel on the behalf of the other
 *	user processes using the filesystem.
 *
 *	The connction between the user process and the kernel is
 *	stateful, so if the connection breaks, Something Bad happens -
 *	I suppose it should be considered a umount - but what about
 *	outstanding writes?
 *
 *	For much the same reasons as nfs and dosfs, there can be no
 *	direct mapping from files, making exec() less memory and time
 *	efficient.  I don't see this as a big problem.
 *
 *	Jeremy Fitzhardinge 1993
 */

#ifndef _LINUX_USERFS_FS
#define _LINUX_USERFS_FS

#include <linux/param.h>	/* For NGROUPS */

/*
 * These operations map closely with the elements in the various
 * operations structures.
 *
 * Things are missing here, or overloaded.  Open does create, mknod and
 * mkdir.  Unlink can also do a rmdir.
 */
typedef enum
{
	up_create,	/* create files, devices, dirs */
	up_lookup,	/* path->file handle */
	up_close,	/* close on fh */
	up_read,	/* read from file */
	up_write,	/* write to file */
	up_truncate,	/* set file length */
	up_fsync,	/* sync file */
	up_readdir,	/* read dir entries */
	up_link,	/* link files */
	up_unlink,	/* unlink file */
	up_symlink,	/* create symlink */
	up_readlink,	/* read link contents */
	up_followlink,	/* resolve path through link */
	up_mount,	/* mount request */
	up_umount,	/* unmount request */
	up_iread,	/* get contents of inode for file */
	up_iwrite,	/* set inode contents */
	up_statfs,	/* stat filesystem wide things */
	up_iput,	/* put an inode after finished */
	up_open,	/* open of a file */
	up_permission,	/* check file permissions */
	up_rename,	/* rename file */
} up_ops;

/* increment this when the structure shapes change */
#define UP_VERSION	14

#include <linux/userfs_types.h>

#ifdef __KERNEL__
typedef unsigned char *(*encode_func)(const void *, unsigned char *);
typedef unsigned char *(*decode_func)(void *, unsigned char *);
typedef unsigned int (*sizeof_func)(const void *);

#define GEN_ENC(var)	(encode_func)(var)
#define GEN_DEC(var)	(decode_func)(var)
#define GEN_SIZ(var)	(sizeof_func)(var)

void userfs_genpkt(struct super_block *, up_preamble *, up_ops);

/* Encode and send request, receive and decode reply */
#define userfs_doop(sb, pre, repl, st, ss, rt, rs)\
	__userfs_doop(sb, pre, repl, \
		      GEN_ENC(encode_##st), (void *)ss, \
		      GEN_DEC(decode_##rt), (void *)rs, \
		      GEN_SIZ(sizeof_##st), GEN_SIZ(sizeof_##rt))

int __userfs_doop(struct super_block *, up_preamble *pre, upp_repl *repl, encode_func, void *, decode_func, void *, sizeof_func, sizeof_func);
#endif /* __KERNEL__ */

#endif /* _LINUX_USERFS_FS */
