/* Mount for people's home dirs */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <utime.h>
#include <dirent.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <getopt.h>

#include <time.h>
#include <pwd.h>

#include <sys/types.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <sys/fcntl.h>

#include <linux/fs.h>
#include <linux/userfs_fs.h>

#include <assert.h>

#include "uio.h"
#include "opnames.h"

struct pwent
{
	uid_t uid;
	gid_t gid;
	int ino;
	char *dir;
	char *name;
	struct passwd pwd;
	struct pwent *nxt;
};

static struct pwent *pwfile;
static int npwent;
static time_t pwdtime;
static char *mpoint;

static void genpkt(upp_repl *repl, int seq, up_ops op)
{
	repl->version = UP_VERSION;
	repl->seq = seq;
	repl->op = op;
	repl->isreq = 0;
	repl->size = 0;
	repl->errno = 0;
}

static void free_pwfile()
{
	struct pwent *nxt;

	for(; pwfile != NULL; pwfile = nxt)
	{
		nxt = pwfile->nxt;
		free(pwfile->dir);
		free(pwfile->name);
		free(pwfile);
	}
}

static void load_pwfile()
{
	struct pwent *pwe;
	struct passwd *pw;
	int ino = 0;
	struct stat st;

	stat("/etc/passwd", &st);

	if (st.st_mtime <= pwdtime)
		return;

	free_pwfile();

	while(pw = getpwent())
	{
		pwe = malloc(sizeof(*pwe));
		pwe->dir = strdup(pw->pw_dir);
		pwe->name = strdup(pw->pw_name);
		pwe->uid = pw->pw_uid;
		pwe->gid = pw->pw_gid;
		pwe->ino = ino++;
		pwe->nxt = pwfile;
		pwfile = pwe;
	}
	npwent = ino;
	endpwent();	
	pwdtime = st.st_mtime;
}

static struct pwent *lookup_pwfile(int ino)
{
	struct pwent *pwe = pwfile;

	for(; pwe != NULL; pwe = pwe->nxt)
		if (pwe->ino == ino)
			return pwe;
	return NULL;
}

#define DIRINO	65535

int do_iread(unsigned char *buf, upp_repl *repl)
{
	upp_iread_s rcv;
	upp_iread_r snd;
	struct pwent *pw;
	time_t t;

	decode_upp_iread_s(&rcv, buf);

	memset(&snd, 0, sizeof(snd));

	DB(printf("ino=%x ", rcv.handle.handle));

	if (rcv.handle.handle == DIRINO)
	{
		snd.ino.mode = S_IFDIR | 0555;
		snd.ino.nlink = 2+npwent;
		t = time(0);
	}
	else if ((pw = lookup_pwfile(rcv.handle.handle)) != NULL)
	{
		snd.ino.mode = S_IFLNK | 0666;
		snd.ino.uid = pw->uid;
		snd.ino.gid = pw->gid;
		snd.ino.size = strlen(pw->dir);
		snd.ino.nlink = 1;
		t = pwdtime;
	} else
		return ENOENT;
	
	snd.handle = rcv.handle;
	snd.ino.mtime = snd.ino.ctime = t;
	snd.ino.atime = time(0);

	repl->size = encode_upp_iread_r(&snd, buf) - buf;

	return 0;
}

int do_readdir(unsigned char *buf, upp_repl *repl)
{
	upp_readdir_s rcv;
	upp_readdir_r snd;
	struct pwent *pwe = pwfile;
	int i;

	decode_upp_readdir_s(&rcv, buf);

	snd.off = 1;
	snd.name.nelem = 0;

	if (rcv.dir.handle != DIRINO)
		return ENOTDIR;

	if (rcv.off == 0)
	{
		snd.name.elems = ".";
		snd.name.nelem = 1;
		snd.file.handle = DIRINO;
	} else if (rcv.off == 1)
	{
		snd.name.elems = "..";
		snd.name.nelem = 2;
		snd.file.handle = DIRINO;
	} else
	{
		rcv.off -= 2;
		snd.off = 0;
		for(i = 0; pwe != NULL && i < rcv.off; pwe = pwe->nxt, i++)
			;


		if (pwe != NULL)
		{
			snd.off = 1;
			snd.name.elems = pwe->name;
			snd.name.nelem = strlen(snd.name.elems);
			snd.file.handle = pwe->ino;
		}
	}

	DB(if (snd.name.nelem) printf("n=%s ", snd.name.elems));
	repl->size = encode_upp_readdir_r(&snd, buf) - buf;

	return 0;
}

int do_lookup(unsigned char *buf, upp_repl *repl)
{
	upp_lookup_s rcv;
	upp_lookup_r snd;
	struct pwent *pw = pwfile;
	
	decode_upp_lookup_s(&rcv, buf);

	if (rcv.dir.handle != DIRINO)
		return ENOTDIR;

	DB(printf("n=%.*s ", rcv.name.nelem, rcv.name.elems));

	if ((rcv.name.nelem == 0) ||
	    (rcv.name.nelem == 1 && *rcv.name.elems == '.') ||
	    (rcv.name.nelem == 2 &&
            	rcv.name.elems[0] == '.' && rcv.name.elems[1] == '.'))
	{
		snd.handle.handle = DIRINO;
	}
	else
	{
		for(pw = pwfile; pw != NULL; pw = pw->nxt)
			if (strncmp(pw->name, rcv.name.elems, rcv.name.nelem) == 0)
				break;

		FREE(rcv.name.elems);

		if (pw == NULL)
			return ENOENT;

		snd.handle.handle = pw->ino;
	}

	repl->size = encode_upp_lookup_r(&snd, buf) - buf;

	return 0;
}

int do_readlink(unsigned char *buf, upp_repl *repl)
{
	upp_readlink_s rcv;
	upp_readlink_r snd;
	struct pwent *pw;

	decode_upp_readlink_s(&rcv, buf);

	pw = lookup_pwfile(rcv.link.handle);

	if (pw == NULL)
		return ENOENT;

	snd.name.elems = pw->dir;
	snd.name.nelem = strlen(pw->dir);

	DB(printf("l=%s ", snd.name.elems));

	repl->size = encode_upp_readlink_r(&snd, buf) - buf;

	return 0;
}

int do_followlink(unsigned char *buf, upp_repl *repl)
{
	upp_followlink_s rcv;
	upp_followlink_r snd;
	struct pwent *pw;

	decode_upp_followlink_s(&rcv, buf);

	pw = lookup_pwfile(rcv.link.handle);

	if (pw == NULL)
		return ENOENT;

	snd.path.elems = pw->dir;
	snd.path.nelem = strlen(pw->dir);

	DB(printf("l=%s ", snd.path.elems));

	repl->size = encode_upp_followlink_r(&snd, buf) - buf;

	return 0;
}

int do_mount(unsigned char *buf, upp_repl *repl)
{
	upp_mount_r snd;

	snd.root.handle = DIRINO;

	repl->size = encode_upp_mount_r(&snd, buf) - buf;

	load_pwfile();

	repl->size = encode_upp_mount_r(&snd, buf) - buf;

	return 0;
}

void run(int wr, int rd)
{
	up_preamble     pre;
	upp_repl         repl;
	unsigned char	buf[4096];
	int error;

	while(readpkt(rd, &pre, buf) == 1)
	{

		DB(printf("Got %s (%d):", opnames[pre.op], pre.op));
		DB(if (pre.isreq == UP_ENQ) printf("enquiry"));

		genpkt(&repl, pre.seq, pre.op);

#define DO(op, b, r)	(pre.isreq == UP_ENQ ? 0 : do_##op(b, r))

		switch(pre.op)
		{
		case up_mount:
			error = DO(mount, buf, &repl);
			break;

		case up_iread:
			error = DO(iread, buf, &repl);
			break;		

		case up_readdir:
			error = DO(readdir, buf, &repl);
			break;
			
		case up_lookup:
			error = DO(lookup, buf, &repl);
			break;

		case up_readlink:
			error = DO(readlink, buf, &repl);
			break;

		case up_followlink:
			error = DO(followlink, buf, &repl);
			break;

#undef DO
		default:
			error = ENOSYS;
			break;
		}
		
		DB(putchar('\n'));

		repl.errno = error;
		if (error != 0)
		{
			DB(printf("replying with errno = \"%s\" (%d)\n",
				strerror(error), error));
			repl.size = 0;
		}

		if (writepkt(wr, &repl, buf) != 1)
			break;
	}
}
		
int
main(int argc, char *argv[])
{
	int infd = 0;
	int outfd = 1;
	int c;

	while((c = getopt(argc, argv, "i:o:m:")) != EOF)
		switch(c)
		{
		case 'i':
			infd = atoi(optarg);
			break;

		case 'o':
			outfd = atoi(optarg);
			break;

		case 'm':
			mpoint = optarg;
			break;
			
		default:
			fprintf(stderr, "%s: bad option\n", argv[0]);
			exit(1);
		}

	signal(SIGCHLD, SIG_IGN);
	
	DB(printf("out=%d in=%d\n", outfd, infd));
	run(outfd, infd);

	DB(printf("done protocol\n"));

	exit(0);
}
