// -*- c++ -*-
// Inode handling classes

#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <linux/fs.h>
#include <linux/userfs_fs.h>

#include <assert.h>

#include "itab.h"

extern "C" {
#include "uio.h"
}

ient::ient(const char *n, hand_t h, ient *d, ient *nxt, itab *tab)
{
	elem = strdup(n);
	handle = h;
	dir = d;
	insert(nxt);

	if (dir != NULL)
		dir->used++;

	opened = 0;
	unlinked = 0;
	fd = -1;
	
	table = tab;

	goo = (unsigned)-1;
}

ient::~ient()
{
	if (used != 0)
		fprintf(stderr, "deleting directory in use!\n");

	remove();
	if (isopen())
		fprintf(stderr, "deleting open file\n");

	opened = 1;
	close();
	free(elem);
	if (dir != NULL)
		dir->used--;
}

void ient::remove()
{
	if (next)
		next->prev = prev;
	if (prev)
		prev->next = next;
	next = prev = NULL;
}

void ient::insert(ient *nxt)
{
	if (next == nxt)
		return;

	if (nxt)
	{
		prev = nxt->prev;
		nxt->prev = this;
	}
	else
		prev = NULL;
	next = nxt;

	assert(next != this);
	assert(prev != this);
}

void itab::use(ient *ent)
{
	if (ent == list)
		return;

	ent->remove();
	ent->insert(list);
	list = ent;
}

int ient::rename(ient *ndir, const char *name)
{
	char toname[1024], *tne;
	char fromname[1024];

	if (dir == NULL || ndir == NULL)
		return EPERM;

	tne = dir->catname(toname);
	catname(fromname);

	*tne++ = '/';
	strcpy(tne, name);
       
	if (::rename(fromname, toname) == -1)
		return errno;

	free(elem);
	elem = strdup(name);
	dir->used--;
	dir = ndir;
	dir->used++;
	return 0;
}

int itab::unlink(ient *ino)
{
	char name[1024];
	int ret;

	if (!ino->isopen())
	{
		struct stat st;

		DB(fputs(" D!", stdout));
		ino->catname(name);
		if (::lstat(name, &st) == -1)
			return errno;

		if (S_ISDIR(st.st_mode))
		{
			if(::rmdir(name) == -1)
				return errno;
		}
		else
			if (::unlink(name) == -1)
				return errno;

		if (list == ino)
			list = ino->next;
		delete ino;
		return 0;
	}

	DB(fputs("d-", stdout));
	ino->unlinked = 1;
	sprintf(name, ".userfs.%x", ino->gethand());
	if ((ret = ino->rename(ino->dir, name)) != 0)
		return ret;

	DB(printf(" tn=%s", name));
	return 0;
}

char * ient::catname(char *buf)
{
	char *where = buf;
	
	if (dir != NULL)
		where = dir->catname(buf);

	if (*elem != '/')
		*where++ = '/';

	strcpy(where, elem);
	where += strlen(elem);
	*where = '\0';
	return where;
}

int ient::open()
{
	char name[1024];
	struct stat st;
	
	if (opened != 0)
	{
		assert(fd != -1);
		goto out;
	}

	catname(name);

	if (stat(name, &st) == -1)
	{
		FAIL("stat");
		DB(printf("filename=%s ", name));
		fd = -1;
		goto out;
	}
	
	fd = ::open(name, S_ISDIR(st.st_mode) ? O_RDONLY: O_RDWR);

 out:
	if (fd != -1)
		opened++;
	return fd;
}

void ient::close()
{
	if (opened == 0)
	{
		assert(fd == -1);
		DB(fputs(" c!", stdout));
		return;
	}

	if (opened > 1)
	{
		opened--;
		assert((signed)opened >= 0);
		DB(fputs(" c-", stdout));
		return;
	}

	if (::close(fd) != -1)
		DB(fputs(" c.", stdout));
	opened--;
	fd = -1;
	assert(opened == 0);

	if (unlinked)
	{
		DB(fputs("d! ", stdout));
		table->unlink(this);
	}
}

int ient::lstat(struct stat *st)
{
	char name[4096];

	catname(name);

	if (::lstat(name, st) == -1)
		return -1;
	return 0;
}

int ient::stat(struct stat *st)
{
	char name[4096];

	catname(name);

	if (::stat(name, st) == -1)
		return -1;
	return 0;
}

ient *itab::map_inode(const char *name, hand_t handle, ient *dir)
{
	ient *ie = get_inode(handle);

	if (ie != NULL)
		return ie;

	list = new ient(name, handle, dir, list, this);
	inodes++;

	DB(printf(" m%x->d%x ", handle, dir->handle));
	return list;
}

ient *itab::get_inode(hand_t handle)
{
	ient *il = list;

	while(il)
	{
		if (il->gethand() == handle)
		{
			use(il);
			return il;
		}
		il = il->next;
	}

	return NULL;
}
