/*
 * Simple mount interface for userfs.  This does permissions checking
 * and allows non-suid processes to be mounted.
 *
 * It checks the proposed mount dir, does the mount and starts up
 * the filesystem program attached to the kernel with pipes to
 * deal with the filesystem operations.  The filesystem process
 * runs as the invoking user.
 *
 * The /etc/mtab file is also updated correctly, so that umount should
 * be able to unmount the filesystem.
 *
 * The input and output fds are inhereted by the filesystem process
 * and are passed on the command-line with the -i fd (kernel to process)
 * and -o fd (process to kernel) options.
 *
 * Usage: muserfs program mount-point [args...]
 *
 * This program is distributed under the terms of the
 * Free Software Foundation General Public Licence.
 * Copyright Jeremy Fitzhardinge 1993
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <mntent.h>
#include <pwd.h>
#include <getopt.h>
#include <signal.h>
#include <fcntl.h>

#include <sys/wait.h>
#include <sys/types.h>
#include <sys/mount.h>
#include <sys/stat.h>

#include <linux/fs.h>
#include <linux/userfs_mount.h>

char *progname = NULL;

__volatile__
void sysfatal(const char *str, ...)
{
	va_list va;
	int er = errno;
	
	va_start(va, str);
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, str, va);
	fprintf(stderr, ": %s\n", strerror(er));
	va_end(va);
	exit(1);
}

void *xmalloc(size_t s)
{
	void *m = malloc(s);

	if (m == NULL)
		sysfatal("memory allocation failure");

	return m;
}

char *fullpath(const char *p)
{
	char cwd[1024];
	int dl;
	char *np;
	
	if (*p == '/')
		return strdup(p);

	if (getcwd(cwd, sizeof(cwd)) == NULL)
		sysfatal("Can't get cwd");

	dl = strlen(cwd);
	np = xmalloc(dl + strlen(p) + 2);

	strcpy(np, cwd);
	np[dl] = '/';
	strcpy(&np[dl]+1, p);

	return np;
}

char *makeopts(char **av, int narg, const char *name, int pid)
{
	int i, j, nblen;
	char *op, *cp;
	char nbuf[100];
	
	sprintf(nbuf, "user=%s,pid=%d,nodev,nosuid,", name, pid);
	nblen = j = strlen(nbuf);
	
	for(i = 0; i < narg; i++)
		j += strlen(av[i]) + 1;

	op = xmalloc(j + 1);

	strcpy(op, nbuf);
	cp = &op[nblen];
	
	for(i = 0; i < narg; i++)
	{
		strcpy(cp, av[i]);
		cp += strlen(av[i]);
		*cp++ = ',';
	}

	cp[-1] = 0;

	return op;
}

void domount(int infd, int outfd, const char *mpoint, char *prog,
	     char **args, int nargs, int async, int kid)
{
	struct userfs_mount mnt;
	struct mntent ment;
	FILE *mtab;
	struct passwd *pwd;
	char *mdev;
	int fd;
	
	if ((pwd = getpwuid(getuid())) == NULL)
		sysfatal("can't get password file entry for uid %d",
			 getuid());
	
	mnt.tokern = infd;
	mnt.fromkern = outfd;
	mnt.version = USERFS_VERSION;

	if (mount(NULL, mpoint, "userfs",
		  MS_MGC_VAL|MS_NODEV|MS_NOSUID, &mnt) == -1)
		sysfatal("mount of %s on %s failed",
			 prog, mpoint);

	close(infd);
	close(outfd);

	mdev = xmalloc(strlen(prog) + 7);
	sprintf(mdev, "%s:%d", prog, kid);
	
	ment.mnt_fsname = mdev;
	ment.mnt_dir = fullpath(mpoint);
	ment.mnt_type = "userfs";
	ment.mnt_opts = makeopts(args, nargs, pwd->pw_name, kid);
	ment.mnt_freq = 0;
	ment.mnt_passno= 0;

	if ((fd = open(MOUNTED"~", O_RDWR|O_CREAT|O_EXCL, 0600)) == -1)
		sysfatal("Can't get "MOUNTED"~ lock file");
	close(fd);
	
	if ((mtab = setmntent(MOUNTED, "a+")) == NULL)
		sysfatal("Can't open " MOUNTED);

	if (addmntent(mtab, &ment) == 1)
		sysfatal("Can't write mount entry");

	if (fchmod(fileno(mtab), 0644) == -1)
		sysfatal("Can't set perms on "MOUNTED);
		
	endmntent(mtab);

	if (unlink(MOUNTED"~") == -1)
		sysfatal("Can't remove "MOUNTED"~");
	
	free(mdev);
	
	if (async)
		return;

	while(wait((int *)0) != kid)
	{
		if (kid == -1)
			sysfatal("wait for pid %d failed", kid);
	}

	/* Use umount to deal with /etc/mtab, etc */
	execl("/sbin/umount", "umount", ment.mnt_dir, NULL);
	sysfatal("exec of \"umount %s\" failed", ment.mnt_dir);
}

int mountok(struct stat *st)
{
	if (!S_ISDIR(st->st_mode))
	{
		errno = ENOTDIR;
		return -1;
	}
	
	if (getuid() != 0 &&
	    (getuid() != st->st_uid ||
	     (st->st_mode & S_IRWXU) != S_IRWXU)
	   )
	{
		errno = EPERM;
		return -1;
	}

	return 0;
}

int main(int argc, char *argv[])
{
	struct stat st;
	int kid, i, ch;
	int p1[2], p2[2];
	char **args;
	char b1[10], b2[10];
	int async = 0;
	
	progname = argv[0];

	if (geteuid() != 0 || getegid() != getgid())
	{
		fprintf(stderr, "%s: must be installed suid root\n",
			progname);
		exit(1);
	}

	for(i = 0; (ch = getopt(argc, argv, "a")) != EOF;)
		switch(ch)
		{
		case 'a':
			async = 1;
			break;
		default:
			i++;
		}

	argc -= optind;
	argv = &argv[optind];
	
	if (i != 0 || argc < 2)
	{
		fprintf(stderr, "Usage: %s program mount-point [args...]\n",
			progname);
		exit(1);
	}

	if (async && getuid() != 0)
	{
		fprintf(stderr, "%s: async (-a) ignored: you can't unmount if you arn't root\n",
			progname);
		async = 0;
	}
	
	if (stat(argv[1], &st) == -1)
		sysfatal("Stat of %s failed", argv[1]);

	if (mountok(&st) == -1)
		sysfatal("can't mount on %s", argv[1]);

	if (pipe(p1) == -1 || pipe(p2) == -1)
		sysfatal("can't create pipe");
	
	switch(kid = fork())
	{
	case -1:	/* fail */
		sysfatal("fork failed");

	case 0:		/* child */
		break;

	default:	/* parent */
		signal(SIGINT, SIG_IGN);
		signal(SIGHUP, SIG_IGN);
		signal(SIGTERM, SIG_IGN);
		
		close(p1[1]);
		close(p2[0]);

		domount(p1[0], p2[1], argv[1], argv[0], 
			&argv[2], argc-2, async, kid);
		exit(0);
	}

	/* child */
	close(p1[0]);
	close(p2[1]);
	
	if (setreuid(-1, getuid()) == -1)
		sysfatal("Can't drop suid permissions");

	
	args = (char **)xmalloc(sizeof(char *)* (argc + 3));

	sprintf(b1, "-i%d", p2[0]);
	sprintf(b2, "-o%d", p1[1]);
	
	args[0] = argv[0];
	args[1] = b1;
	args[2] = b2;
	args[3] = "-m";
	args[4] = argv[1];

	for(i = 0; i <= argc-3; i++)
		args[i+5] = argv[i+2];
	args[i+5] = NULL;
	
	execvp(args[0], args);

	sysfatal("exec of %s failed", args[0]);
}

       
