// -*- c++ -*-
/*
 * Test user process
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <utime.h>
#include <dirent.h>
#include <errno.h>
#include <signal.h>
#include <string.h>

#include <sys/types.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <sys/fcntl.h>

extern "C"
{
#include <linux/fs.h>
#include <linux/userfs_fs.h>
#include <linux/userfs_mount.h>
}

#include <assert.h>

#include "itab.h"
#include "uio.h"
#include "opnames.h"

inline int min(int x, int y)
{
	return x < y ? x : y;
}

void *xmalloc(size_t sz)
{
	void *ptr = malloc(sz);

	if (ptr == NULL)
	{
		fprintf(stderr, "out of memory\n");
		exit(1);
	}

	return ptr;
}

char *mpoint;

static void genpkt(upp_repl *repl, int seq, up_ops op)
{
	repl->version = UP_VERSION;
	repl->seq = seq;
	repl->op = op;
	repl->isreq = 0;
	repl->size = 0;
	repl->errno = 0;
}

void quit(int sig)
{
	fprintf(stderr, "unmounting %s\n", mpoint);

	signal(sig, quit);

	switch(fork())
	{
	case 0:
		if (umount(mpoint) == -1)
			perror("umount");
		exit(0);
	case -1:
		perror("quitfork");
		break;
	}
}

itab itable;

extern int do_readdir(unsigned char *, upp_repl *);

int do_lookup(unsigned char *buf, upp_repl *repl)
{
	DIR *dir = NULL;
	struct dirent *de;
	int ret = 0, needfree = 1;
	char name[1024];
	struct stat st;
	hand_t handle;
	ient *diri;
	ient *ino;
	upp_lookup_s rcv;
	upp_lookup_r snd;

	decode_upp_lookup_s(&rcv, buf);

	if ((diri = itable.get_inode(rcv.dir.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad inode/handle %x\n",
			__FILE__, __LINE__, rcv.dir.handle);
		return EINVAL;
	}

	diri->catname(name);

	DB(printf("d=%s, n=%.*s", name, rcv.name.nelem, rcv.name.elems));

	if ((dir = opendir(name)) == NULL)
	{
		ret = errno;
		FAIL("opendir");
		goto out;
	}

	if (rcv.name.nelem == 0)
	{
		FREE(rcv.name.elems);
		needfree = 0;
		rcv.name.elems = ".";
		rcv.name.nelem = 1;
	}

	while((de = readdir(dir)) != NULL)
		if (strncmp(de->d_name, rcv.name.elems, rcv.name.nelem) == 0)
			break;

	if (de == NULL)
	{
		ret = ENOENT;
		goto out;
	}

	strcat(name, "/");
	strcat(name, de->d_name);
	if (lstat(name, &st) == -1)
		FAIL("lstat");
	handle = itable.genhand(st.st_dev, st.st_ino);
	
	if ((ino = itable.map_inode(de->d_name, handle, diri)) == NULL)
	{
		ret = EMFILE;
		goto out;
	}

	DB(printf(", h=%x", handle));

	snd.handle.handle = handle;

	repl->size = encode_upp_lookup_r(&snd, buf) - buf;

	if (!ino->isopen())
		ino->open();

	ret = 0;
 out:
	if (needfree)
		FREE(rcv.name.elems);

	if (dir != NULL)
		closedir(dir);

	return ret;
}

int do_iread(unsigned char *buf, upp_repl *repl)
{
	struct stat st;
	ient *ino;
	upp_iread_s rcv;
	upp_iread_r snd;
	up_inode *pi = &snd.ino;

	decode_upp_iread_s(&rcv, buf);

	if ((ino = itable.get_inode(rcv.handle.handle)) == NULL)
	{
		fprintf(stderr, "bad inode/handle %x ", rcv.handle.handle);
		return EINVAL;
	}

	if (ino->lstat(&st) == -1)
	{
		FAIL("lstat");
		return errno;
	}

	snd.handle = rcv.handle;
	pi->mode = st.st_mode;
	pi->nlink = st.st_nlink;
	pi->uid = st.st_uid;
	pi->gid = st.st_gid;
	pi->size = st.st_size;
	pi->atime = st.st_atime;
	pi->mtime = st.st_mtime;
	pi->ctime = st.st_ctime;
	pi->blksize = st.st_blksize;
	pi->blocks = st.st_blocks;
	pi->rdev = st.st_rdev;

	if (!ino->isopen())
		ino->open();

	DB(printf("ino %x ", rcv.handle.handle));

	repl->size = encode_upp_iread_r(&snd, buf) - buf;

	return 0;
}

int do_iwrite(unsigned char *buf, upp_repl *repl)
{
	int fd;
	struct utimbuf utbuf;
	char name[4096];
	ient *ino;
	struct stat st;
	upp_iwrite_s rcv;

	decode_upp_iwrite_s(&rcv, buf);
	
	if ((ino = itable.get_inode(rcv.handle.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad inode/handle %x\n",
			__FILE__, __LINE__, rcv.handle.handle);
		return EINVAL;
	}

	DB(printf("writing ino %x", rcv.handle.handle));

	if (ino->lstat(&st) == -1)
		FAIL("fstat");
	if (S_ISLNK(st.st_mode))
		return 0;

	fd = ino->open();

	if (fchmod(fd, rcv.ino.mode) == -1)
		FAIL("fchmod");
	if (fchown(fd, rcv.ino.uid, rcv.ino.gid) == -1)
		FAIL("fchown");
	if (fstat(fd, &st) == -1)
		FAIL("fstat");
	if (S_ISREG(st.st_mode) && ftruncate(fd, rcv.ino.size) == -1)
		FAIL("ftruncate");

	ino->close();

	utbuf.actime = rcv.ino.atime;
	utbuf.modtime = rcv.ino.mtime;
	
	ino->catname(name);

	if (utime(name, &utbuf) == -1)
		FAIL("utime");
	
	return 0;
}

int do_iput(unsigned char *buf, upp_repl *repl)
{
	ient *ino;
	upp_iput_s rcv;

	decode_upp_iput_s(&rcv, buf);
	
	if ((ino = itable.get_inode(rcv.handle.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad inode/handle %x\n",
			__FILE__, __LINE__, rcv.handle.handle);
		return EINVAL;
	}

	DB(printf("putting %x", rcv.handle.handle));
	
	ino->close();

	return 0;
}

int do_read(unsigned char *mbuf, upp_repl *repl)
{
	ient *ino;
	int fd, err;
	size_t sz, rd;
	char *buf;
	upp_read_s rcv;
	upp_read_r snd;
	unsigned int max_size;

	snd.data.nelem = 0;
	max_size = 4096 - sizeof_upp_read_r(&snd);
	
	decode_upp_read_s(&rcv, mbuf);
	
	if ((ino = itable.get_inode(rcv.file.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad inode/handle %x\n",
			__FILE__, __LINE__, rcv.file.handle);
		return EINVAL;
	}

	fd = ino->open();

	if (lseek(fd, rcv.off, SEEK_SET) == -1)
	{
		err = errno;
		FAIL("lseek");
		return err;
	}

	sz = min(max_size, rcv.size);
	buf = (char *)xmalloc(sz);
	
	if ((rd = read(fd, buf, sz)) == -1)
	{
		err = errno;
		FAIL("read");
		free(buf);
		return err;
	}

	ino->close();

	if (rd == 0)
	{
		free(buf);
		buf = NULL;
	}

	snd.data.elems = (unsigned char *)buf;
	snd.data.nelem = rd;

	DB(printf(" read %d bytes ", rd));

	repl->size = encode_upp_read_r(&snd, mbuf) - mbuf;

	if (buf)
		free(buf);
	
	return 0;
}

int do_write(unsigned char *buf, upp_repl *repl)
{
	ient *ino;
	int fd, err;
	size_t wr;
	upp_write_s rcv;
	upp_write_r snd;

	decode_upp_write_s(&rcv, buf);
	
	if ((ino = itable.get_inode(rcv.file.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad inode/handle %x\n",
			__FILE__, __LINE__, rcv.file.handle);
		return EINVAL;
	}

	fd = ino->open();

	if (lseek(fd, rcv.off, SEEK_SET) == -1)
	{
		err = errno;
		FAIL("lseek");
		return err;
	}

	if ((wr = write(fd, rcv.data.elems, rcv.data.nelem)) == -1)
	{
		err = errno;
		FAIL("write");
		return err;
	}

	ino->close();

	snd.wrote = wr;

	repl->size = encode_upp_write_r(&snd, buf) - buf;
	
	DB(printf(" wrote %d bytes ", wr));

	return 0;
}

int do_create(unsigned char *buf, upp_repl *repl)
{
	ient *dir;
	char path[1024];
	char *pp;
	struct stat st;
	int ret;
	upp_create_s rcv;
	upp_create_r snd;
	
	decode_upp_create_s(&rcv, buf);
	
	if ((dir = itable.get_inode(rcv.dir.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad inode/handle %x\n",
			__FILE__, __LINE__, rcv.dir.handle);
		return EINVAL;
	}

	dir->catname(path);
	pp = &path[strlen(path)];

	*pp++ = '/';
	memcpy(pp, rcv.name.elems, rcv.name.nelem);
	pp[rcv.name.nelem] = 0;

	FREE(rcv.name.elems);
	
	ret = -1;

	DB(printf("mode=%o ", rcv.mode));

	if (S_ISREG(rcv.mode))
	{
		ret = creat(path, rcv.mode);
		if (ret >= 0)
		{
			close(ret);
			ret = 0;
		}
	}
	else if (S_ISDIR(rcv.mode))
		ret = mkdir(path, rcv.mode);
	else if (S_ISBLK(rcv.mode) || S_ISCHR(rcv.mode))
	{
		DB(printf("dev=%x ", rcv.rdev));
		ret = mknod(path, rcv.mode, rcv.rdev);
	}
	else
		ret = EINVAL;

	if (ret == 0)
	{
		ient *file;

		fchown(ret, rcv.cred.euid, rcv.cred.egid);
		fchmod(ret, rcv.mode);
		if (stat(path, &st) == -1)
			FAIL("stat");

		file = itable.map_inode(pp, itable.genhand(st.st_dev, st.st_ino), dir);
		snd.file.handle = file->gethand();
	}

	if (ret == 0)
		repl->size = encode_upp_create_r(&snd, buf) - buf;
	
	return ret;
}

int do_unlink(unsigned char *buf, upp_repl *repl)
{
	ient *dir;
	ient *file;
	char name[1024], *ep;
	struct stat st;
	int ret;
	upp_unlink_s rcv;

	decode_upp_unlink_s(&rcv, buf);
	
	if ((dir = itable.get_inode(rcv.dir.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad inode/handle %x\n",
			__FILE__, __LINE__, rcv.dir.handle);
		return EINVAL;
	}

	ep = dir->catname(name);
	*ep++ = '/';
	memcpy(ep, rcv.name.elems, rcv.name.nelem);
	ep[rcv.name.nelem] = '\0';
	DB(printf("d=%x n=%s ", dir->gethand(), name));

	FREE(rcv.name.elems);
	
	if (lstat(name, &st) == -1)
	{
		ret = errno;
		FAIL("lstat");
		return ret;
	}
	
	if ((file = itable.map_inode(ep, itable.genhand(st.st_dev, st.st_ino), dir)) == NULL)
		return EMFILE;

	DB(printf("f=%x %d", file->gethand(), file->isopen()));

	if ((ret = itable.unlink(file)) != 0)
		fprintf(stderr, "do_unlink: unlink %s failed with %d\n", name, ret);

	return ret;
}

int do_readlink(unsigned char *buf, upp_repl *repl)
{
	ient *lnk;
	struct stat st;
	char name[1024];
	char lname[1024];
	int len;
	upp_readlink_s rcv;
	upp_readlink_r snd;

	decode_upp_readlink_s(&rcv, buf);
	
	if ((lnk = itable.get_inode(rcv.link.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad inode/handle %x\n",
			__FILE__, __LINE__, rcv.link.handle);
		return EINVAL;
	}

	lnk->lstat(&st);

	if (!S_ISLNK(st.st_mode))
		return EINVAL;

	lnk->catname(lname);

	if ((len = readlink(lname, name, 1024)) == -1)
	{
		int ret = errno;
		FAIL("readlink");
		return ret;
	}

	snd.name.nelem = len;
	snd.name.elems = name;

	repl->size = encode_upp_readlink_r(&snd, buf) - buf;

	DB(printf("l=%s ", name));

	return 0;
}

int do_followlink(unsigned char *buf, upp_repl *repl)
{
	ient *lnk;
	ient *dir;
	struct stat st;
	char lpath[1024];
	upp_followlink_s rcv;
	upp_followlink_r snd;

	decode_upp_followlink_s(&rcv, buf);
	
	DB(printf("flag=%d mode=%d ", rcv.flag, rcv.mode));

	if ((lnk = itable.get_inode(rcv.link.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad link inode/handle %x\n",
			__FILE__, __LINE__, rcv.link.handle);
		return EINVAL;
	}

	if ((dir = itable.get_inode(rcv.dir.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad dir inode/handle %x\n",
			__FILE__, __LINE__, rcv.dir.handle);
		return EINVAL;
	}

	if (lnk->lstat(&st) == -1)
		FAIL("lnk->lstat");

	lnk->catname(lpath);
	if (!S_ISLNK(st.st_mode))
		return EINVAL;

	if (readlink(lpath, lpath, sizeof(lpath)) == -1)
		return errno;

	snd.path.nelem = strlen(lpath);
	snd.path.elems = lpath;
	repl->size = encode_upp_followlink_r(&snd, buf) - buf;
	
	return 0;
}

int do_symlink(unsigned char *buf, upp_repl *repl)
{
	upp_symlink_s rcv;
	ient *dir;
	int ret;
	char path[1024], *pp;
	char slink[1024];
	
	decode_upp_symlink_s(&rcv, buf);

	if ((dir = itable.get_inode(rcv.dir.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad dir inode/handle %x\n",
			__FILE__, __LINE__, rcv.dir.handle);
		ret = EINVAL;
		goto out;
	}

	pp = dir->catname(path);
	*pp++ = '/';
	memcpy(pp, rcv.name.elems, rcv.name.nelem);
	pp[rcv.name.nelem] = 0;

	memcpy(slink, rcv.symname.elems, rcv.symname.nelem);
	slink[rcv.symname.nelem] = 0;

	if ((ret = symlink(slink, path)) == -1)
		FAIL("symlink");
		
 out:
	if (rcv.name.nelem)
		FREE(rcv.name.elems);
	if (rcv.symname.nelem)
		FREE(rcv.symname.elems);

	return ret;
}

int do_open(unsigned char *buf, upp_repl *repl)
{
	upp_open_s rcv;
	ient *fil;
	
	decode_upp_open_s(&rcv, buf);

	DB(printf("h=%x ", rcv.file.handle));

	if ((fil = itable.get_inode(rcv.file.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad inode/handle %x\n",
			__FILE__, __LINE__, rcv.file.handle);
		return EINVAL;
	}

	if (!fil->isopen())
		fil->open();

	return 0;
}

int do_close(unsigned char *buf, upp_repl *repl)
{
	upp_close_s rcv;
	ient *fil;
	
	decode_upp_close_s(&rcv, buf);

	DB(printf("h=%x ", rcv.file.handle));

	if ((fil = itable.get_inode(rcv.file.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad inode/handle %x\n",
			__FILE__, __LINE__, rcv.file.handle);
		return EINVAL;
	}

	fil->close();

	return 0;
}

static int in_group_p(gid_t gid, const up_cred *cr)
{
	int i;

	if (gid == cr->egid)
		return 1;
	
	for(i = 0; i < cr->groups.nelem; i++)
		if (gid == cr->groups.elems[i])
			return 1;

	return 0;
}

static int suser(up_cred *cr)
{
	return cr->euid == 0;
}

int do_permission(unsigned char *buf, upp_repl *repl)
{
	upp_permission_s rcv;
	struct stat st;
	ient *fil;
	int mode;
	int mask;
	
	decode_upp_permission_s(&rcv, buf);

	DB(printf("h=%x, uid=%d gid=%d ", rcv.file.handle, rcv.cred.uid, rcv.cred.gid));

	mask = rcv.mask;
	
	if ((fil = itable.get_inode(rcv.file.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad inode/handle %x\n",
			__FILE__, __LINE__, rcv.file.handle);
		return EINVAL;
	}

	fil->stat(&st);

	mode = st.st_mode;
	
	if (rcv.cred.euid == st.st_uid)
		mode >>= 6;
	else if (in_group_p(st.st_gid, &rcv.cred))
		mode >>= 3;
	if (((mode & mask & 0007) == mask) || suser(&rcv.cred))
		return 0;

	return EPERM;
}

#define ROOTDIR "/usr"

int do_mount(unsigned char *buf, upp_repl *repl)
{
	hand_t roothand;
	upp_mount_r	mnt;
	struct stat st;
	int err;
		
	if (stat(ROOTDIR, &st) == -1)
	{
		err = errno;
		FAIL("stat");
		return err;
	}
	roothand = itable.genhand(st.st_dev, st.st_ino);

	if (itable.map_inode(ROOTDIR, roothand, NULL) == NULL)
	{
		FAIL("map_inode");
		return EMFILE;
	}

	mnt.root.handle = roothand;
	repl->size = encode_upp_mount_r(&mnt, buf) - buf;

	return 0;
}

int do_rename(unsigned char *buf, upp_repl *repl)
{
	upp_rename_s rcv;
	ient *odir;
	ient *ndir;
	ient *ofile;
	int ret;
	char name[1024];
	char *pp;
	struct stat st;
	
	decode_upp_rename_s(&rcv, buf);

	if ((odir = itable.get_inode(rcv.odir.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad odir inode/handle %x\n",
			__FILE__, __LINE__, rcv.odir.handle);
		ret = EINVAL;
		goto out;
	}

	if ((ndir = itable.get_inode(rcv.ndir.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad ndir inode/handle %x\n",
			__FILE__, __LINE__, rcv.ndir.handle);
		ret = EINVAL;
		goto out;
	}

	pp = odir->catname(name);
	*pp++ = '/';
	strncpy(pp, rcv.oname.elems, rcv.oname.nelem);
	pp[rcv.oname.nelem] = 0;

	if (lstat(name, &st) == -1)
	{
		ret = errno;
		FAIL("stat");
		goto out;
	}

	if ((ofile = itable.map_inode(pp,
				      itable.genhand(st.st_dev, st.st_ino),
				      odir)) == NULL)
	{
		ret = EMFILE;
		goto out;
	}

	strncpy(name, rcv.nname.elems, rcv.nname.nelem);
	name[rcv.nname.nelem] = 0;
	
	ret = ofile->rename(ndir, name);
	
 out:
	FREE(rcv.oname.elems);
	FREE(rcv.nname.elems);
	
	return ret;
}

void run(int wr, int rd)
{
	up_preamble     pre;
	upp_repl         repl;
	unsigned char	buf[4096];
	int error;

	while(readpkt(rd, &pre, buf) == 1)
	{

		DB(printf("Got %s (%d):", opnames[pre.op], pre.op));
		DB(if (pre.isreq == UP_ENQ) printf("enquiry"));

		genpkt(&repl, pre.seq, pre.op);

#define DO(op, b, r)	(pre.isreq == UP_ENQ ? 0 : do_##op(b, r))

		switch(pre.op)
		{
		case up_mount:
			error = DO(mount, buf, &repl);
			break;

		case up_umount:
			repl.errno = 0;
			error = 0;
			break;
			
		case up_iread:
			error = DO(iread, buf, &repl);
			break;		

		case up_iwrite:
			error = DO(iwrite, buf, &repl);
			break;

		case up_readdir:
			error = DO(readdir, buf, &repl);
			break;
			
		case up_lookup:
			error = DO(lookup, buf, &repl);
			break;

		case up_iput:
			error = DO(iput, buf, &repl);
			break;

		case up_read:
			error = DO(read, buf, &repl);
			break;

		case up_write:
			error = DO(write, buf, &repl);
			break;

		case up_create:
			error = DO(create, buf, &repl);
			break;

		case up_unlink:
			error = DO(unlink, buf, &repl);
			break;

		case up_readlink:
			error = DO(readlink, buf, &repl);
			break;

		case up_symlink:
			error = DO(symlink, buf, &repl);
			break;
#if 0
		case up_followlink:
			error = DO(followlink, buf, &repl);
			break;

		case up_open:
			error = DO(open, buf, &repl);
			break;

		case up_permission:
			error = DO(permission, buf, &repl);
			break;

		case up_close:
			error = DO(close, buf, &repl);
			break;
#endif
		case up_rename:
			error = DO(rename, buf, &repl);
			break;
#undef DO
		default:
			error = ENOSYS;
			break;
		}
		
		DB(putchar('\n'));

		repl.errno = error;
		if (error != 0)
		{
			DB(printf("replying with errno = \"%s\" (%d)\n",
				strerror(error), error));
			repl.size = 0;
		}

		if (writepkt(wr, &repl, buf) != 1)
			break;
	}
}
		
int
main(int argc, char *argv[])
{
	struct userfs_mount mnt;
	int p1[2];	/* From us to kernel */
	int p2[2];	/* From kernel to us */

	if (argc != 2)
	{
		fprintf(stderr, "Usage: %s mount-point\n", argv[0]);
		exit(1);
	}

	if (pipe(p1) == -1)
	{
		perror("pipe 1");
		exit(1);
	}
	if (pipe(p2) == -1)
	{
		perror("pipe 2");
		exit(1);
	}

	mnt.tokern = p1[0];
	mnt.fromkern = p2[1];
	mnt.version = USERFS_VERSION;

	DB(printf("mnt .tokern=%d .fromkern=%d .version=%d\n",
		mnt.tokern, mnt.fromkern, mnt.version));

	mpoint = argv[1];
	signal(SIGINT, quit);
	signal(SIGTERM, quit);
	signal(SIGCHLD, SIG_IGN);

	/*
	 * Fork so we can do protocol during the mount
	 */
	switch(fork())
	{
		int ex;
	case 0:
		ex = 0;
		if (mount(NULL, argv[1], "userfs", MS_MGC_VAL|MS_NODEV|MS_NOSUID, &mnt) == -1)
		{
			perror("mount");
			ex = 1;
		}
		close(p1[0]);
		close(p2[1]);
		exit(ex);
	case -1:
		perror("fork");
		exit(1);
	}

	close(p1[0]);
	close(p2[1]);

	DB(printf("doing protocol, tokern=%d, fromkern=%d\n", p1[1], p2[0]));

	run(p1[1], p2[0]);

	DB(printf("done protocol\n"));

	exit(0);
}
